# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/datacenter.ipynb.

# %% auto 0
__all__ = ['Datacenter_Enum', 'Dataflow_Type_Filter_Enum', 'generate_search_datacenter_body',
           'generate_search_datacenter_account_body', 'search_datacenter', 'get_lineage_upstream']

# %% ../../nbs/routes/datacenter.ipynb 2
from enum import Enum

import httpx

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda

# %% ../../nbs/routes/datacenter.ipynb 3
class Datacenter_Enum(Enum):
    DATASET = "DATASET"
    DATAFLOW = "DATAFLOW"
    ACCOUNT = "ACCOUNT"


class Dataflow_Type_Filter_Enum(Enum):
    ADR = {
        "filterType": "term",
        "field": "data_flow_type",
        "value": "ADR",
        "name": "ADR",
        "not": False,
    }

    MYSQL = {
        "filterType": "term",
        "field": "data_flow_type",
        "value": "MYSQL",
        "name": "MYSQL",
        "not": False,
    }

    REDSHIFT = {
        "filterType": "term",
        "field": "data_flow_type",
        "value": "MYSQL",
        "name": "MYSQL",
        "not": False,
    }

    MAGICV2 = {
        "filterType": "term",
        "field": "data_flow_type",
        "value": "MAGIC",
        "name": "Magic ETL v2",
        "not": False,
    }

    MAGIC = {
        "filterType": "term",
        "field": "data_flow_type",
        "value": "ETL",
        "name": "Magic ETL",
        "not": False,
    }

# %% ../../nbs/routes/datacenter.ipynb 4
def generate_search_datacenter_body(
    entities_list: list[str] = None, # defaults to Dataset
    filters: list[dict] = None,
    combineResults: bool = True,
    count: int = 10,
    offset: int = 0,
):
    entities_list = entities_list or ["DATASET"]
    return {
        "entities": entities_list,
        "filters": filters or [],
        "combineResults": combineResults,
        "query": "*",
        "count": count,
        "offset": offset,
    }

# %% ../../nbs/routes/datacenter.ipynb 5
def generate_search_datacenter_account_body(
    search_str: str, is_exact_match: bool = True
):
    return {
        "count": 100,
        "offset": 0,
        "combineResults": False,
        "query": search_str if is_exact_match else f"*{search_str}*",
        "filters": [],
        "facetValuesToInclude": [
            "DATAPROVIDERNAME",
            "OWNED_BY_ID",
            "VALID",
            "USED",
            "LAST_MODIFIED_DATE",
        ],
        "queryProfile": "GLOBAL",
        "entityList": [["account"]],
        "sort": {"fieldSorts": [{"field": "display_name_sort", "sortOrder": "ASC"}]},
    }

# %% ../../nbs/routes/datacenter.ipynb 6
async def search_datacenter(
    auth: dmda.DomoAuth,
    loop_until_end: bool = True,
    maximum: int = None,
    body: dict = None,
    entity_ls: list[str] = None, # defaults to dataset
    arr_fn: callable = None,
    limit=100,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
) -> rgd.ResponseGetData:

    entity_ls = entity_ls or ["DATASET"]
    if not body:
        body = generate_search_datacenter_body(
            entities_list=entity_ls,
            filters=[],
            combineResults=False,
        )

    if not arr_fn:

        def arr_fn(res):
            return res.response.get("searchObjects")

    url = f"https://{auth.domo_instance}.domo.com/api/search/v1/query"

    res = await gd.looper(
        auth=auth,
        session=session,
        url=url,
        body=body,
        offset_params={"offset": "offset", "limit": "count"},
        arr_fn=arr_fn,
        method="POST",
        maximum = maximum,
        offset_params_in_body=True,
        limit=limit,
        loop_until_end=loop_until_end,
        debug_api=debug_api,
    )

    return res

# %% ../../nbs/routes/datacenter.ipynb 9
async def get_lineage_upstream(
    auth: dmda.DomoAuth,
    entity_type: str,
    entity_id: str,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
):
    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/lineage/{entity_type}/{entity_id}"

    params = {"traverseDown": "false"}

    return await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        params=params,
        session=session,
        debug_api=debug_api,
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/user.ipynb.

# %% auto 0
__all__ = ['get_all_users', 'generate_search_users_body_by_id', 'generate_search_users_body_by_email', 'process_v1_search_users',
           'SearchUser_NoResults', 'search_users', 'search_virtual_user_by_subscriber_instance', 'create_user',
           'set_user_landing_page', 'reset_password', 'request_password_reset']

# %% ../../nbs/routes/user.ipynb 3
import utils.DictDot as dd
import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de


# %% ../../nbs/routes/user.ipynb 5
async def get_all_users(
    auth: dmda.DomoAuth, debug_api: bool = False
) -> rgd.ResponseGetData:

    """retrieves all users from Domo"""
    url = f"https://{ auth.domo_instance}.domo.com/api/content/v2/users"

    return await gd.get_data(url=url, method="GET", auth=auth, debug_api=debug_api)

# %% ../../nbs/routes/user.ipynb 9
def generate_search_users_body_by_id(
    user_ids: list[str],  # list of user ids to search
) -> dict:  # dict to pass to search v1_users_search_api
    """search v1_users_search_api"""

    return {
        # "showCount": true,
        # "count": false,
        "includeDeleted": False,
        "includeSupport": False,
        "filters": [
            {"field": "id", "filterType": "value", "values": user_ids, "operator": "EQ"}
        ],
    }

# %% ../../nbs/routes/user.ipynb 10
def generate_search_users_body_by_email(
    user_email_ls: list[ str ],  # list of user emails to search.  Note:  search does not appear to be case sensitive
) -> dict:  # dict to pass to search v1_users_search_api
    """search v1_users_search_api"""

    return {
        # "showCount": true,
        # "count": false,
        "includeDeleted": False,
        "includeSupport": False,
        "limit": 200,
        "offset": 0,
        "sort": {"field": "displayName", "order": "ASC"},
        "filters": [
            {
                "filterType": "text",
                "field": "emailAddress",
                "text": " ".join(user_email_ls),
            }
        ],
    }

# %% ../../nbs/routes/user.ipynb 11
def process_v1_search_users(v1_user_ls: list[dict] # list of users from v1_users_search API
) -> list[dict]: # sanitized list of users.
    """sanitizes the response from v1_users_search API and removes unecessary attributes"""

    clean_users = []

    for obj_user in v1_user_ls:
        
        dd_user = dd.DictDot(obj_user)

        clean_users.append( {
            "id": dd_user.id,
            "displayName": dd_user.displayName,
            "roleId": dd_user.roleId,
            "userName": dd_user.userName,
            "emailAddress": dd_user.emailAddress,
        })
    
    return clean_users


# %% ../../nbs/routes/user.ipynb 12
class SearchUser_NoResults(de.DomoError):
    def __init__(self, domo_instance,  function_name = "search_users"):
        super().__init__(message = "query returned no users", function_name = function_name)

async def search_users(
    auth: dmda.DomoAuth,
    body: dict,
    debug_api: bool = False,
    return_raw: bool = False
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/identity/v1/users/search"

    res = await gd.get_data(
        url=url,
        method="POST",
        auth=auth,
        body=body,
        debug_api=debug_api,
    )

    if return_raw:
        return res

    if not res.is_success:
        return res

    if res.is_success and len(res.response.get('users')) == 0:
        raise SearchUser_NoResults(domo_instance = auth.domo_instance)
    
    res.response = process_v1_search_users(res.response.get('users'))
    return res



# %% ../../nbs/routes/user.ipynb 16
async def search_virtual_user_by_subscriber_instance(
    auth: dmda.DomoAuth, # domo auth object
    subscriber_instance_ls : list[str], # list of subscriber domo instances
    debug_api: bool = False, # debug API requests
) -> rgd.ResponseGetData: # list of virtual domo users
    """retrieve virtual users for subscriber instances tied to one publisher"""

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/proxy_user/domain/"

    body = {"domains": [f"{subscriber_instance}.domo.com"
                        for subscriber_instance in subscriber_instance_ls]}

    return await gd.get_data(
        url=url,
        method="POST",
        auth=auth,
        body=body,
        debug_api=debug_api,
    )

# %% ../../nbs/routes/user.ipynb 20
async def create_user(
    auth: dmda.DomoAuth,
    display_name: str,
    email: str,
    role_id: int,
    debug_api: bool = False,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/content/v3/users"

    body = {"displayName": display_name, "detail": {"email": email}, "roleId": role_id}

    return await gd.get_data(
        url=url, method="POST", body=body, auth=auth, debug_api=debug_api
    )

# %% ../../nbs/routes/user.ipynb 21
async def set_user_landing_page(
    auth: dmda.DomoAuth, user_id: str, page_id: str, debug_api: bool = False
):

    url = f"https://{auth.domo_instance}.domo.com/api/content/v1/landings/target/DESKTOP/entity/PAGE/id/{page_id}/{user_id}"

    return await gd.get_data(
        url=url,
        method="PUT",
        auth=auth,
        # body = body,
        debug_api=debug_api,
    )

# %% ../../nbs/routes/user.ipynb 22
async def reset_password(
    auth: dmda.DomoAuth,
    user_id: str,
    new_password: str,
    debug_api: bool = False,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/identity/v1/password"

    body = {"domoUserId": user_id, "password": new_password}

    return await gd.get_data(
        url=url,
        method="PUT",
        auth=auth,
        body=body,
        debug_api=debug_api,
    )

# %% ../../nbs/routes/user.ipynb 23
async def request_password_reset(
    domo_instance: str, email: str, locale="en-us", debug_api: bool = False
):
    url = f"https://{domo_instance}.domo.com/api/domoweb/auth/sendReset"

    params = {"email": email, "local": locale}

    return await gd.get_data(
        url=url, method="GET", params=params, auth=None, debug_api=debug_api
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/client/10_get_data.ipynb.

# %% auto 0
__all__ = ['get_data_aiohttp', 'GetData_Error', 'get_data', 'LooperError', 'looper', 'looper_aiohttp']

# %% ../../nbs/client/10_get_data.ipynb 2
from typing import Optional, Union

from pprint import pprint

import httpx
import aiohttp

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoError as de


# %% ../../nbs/client/10_get_data.ipynb 3
async def get_data_aiohttp(
    url: str,
    method: str,
    auth: dmda.DomoAuth,
    content_type: Optional[dict] = None,
    headers: Optional[dict] = None,
    # if no session passed by default will create and close session during execution
    session: Optional[aiohttp.ClientSession] = None,
    body: Union[dict, str, None] = None,
    params: Optional[dict] = None,
    debug_api: bool = False,
    process_stream: bool = False,
    stream_chunks: int = 10
) -> rgd.ResponseGetData:
    """async wrapper for asyncio requests"""

    if auth and not auth.token:
        await auth.get_auth_token()

    if headers is None:
        headers = {}

    is_close_session = False

    if session is None:
        is_close_session = True
        session = session or aiohttp.ClientSession()

    headers = {
        "Content-Type": content_type or "application/json",
        "Connection": "keep-alive",
        "accept": "application/json, text/plain",
        **headers,
    }

    if auth:
        headers.update(**auth.auth_header)

    if debug_api:
        pprint(
            {
                "method": method,
                "url": url,
                "headers": headers,
                "json": body,
                "params": params,
            }
        )

    try:
        if headers.get("Content-Type") == "application/json":
            res = await session.request(
                method=method.upper(),
                url=url,
                headers=headers,
                json=body,
                params=params,
            )

        elif body is not None:
            res = await session.request(
                method=method.upper(),
                url=url,
                headers=headers,
                data=body,
                params=params,
            )

        else:
            res = await session.request(
                method=method.upper(), url=url, headers=headers, params=params
            )
        return await rgd.ResponseGetData._from_aiohttp_response(res, auth=auth, process_stream=process_stream, stream_chunks=stream_chunks)

    except Exception as e:
        print(e)

    finally:
        if is_close_session:
            await session.close()


# %% ../../nbs/client/10_get_data.ipynb 6
class GetData_Error(de.DomoError):
    def __init__(self, message, url):
        super().__init__( message = message, domo_instance = url)
        

# %% ../../nbs/client/10_get_data.ipynb 7
async def get_data(
    url: str,
    method: str,
    auth: dmda.DomoAuth,
    content_type: Optional[dict] = None,
    headers: Optional[dict] = None,
    body: Union[dict, str, None] = None,
    params: Optional[dict] = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
    is_follow_redirects: bool = False,
    timeout = 5
) -> rgd.ResponseGetData:
    """async wrapper for asyncio requests"""

    if debug_api:
        print("🐛 debugging get_data")

    if auth and not auth.token:
        await auth.get_auth_token()

    if headers is None:
        headers = {}

    headers = {
        "Content-Type": content_type or "application/json",
        "Connection": "keep-alive",
        "accept": "application/json, text/plain",
        **headers,
    }

    if auth:
        headers.update(**auth.auth_header)

    if debug_api:
        pprint(
            {
                "method": method,
                "url": url,
                "headers": headers,
                "body": body,
                "params": params,
            }
        )

    is_close_session = False if session else True

    session = session or httpx.AsyncClient()


    attempt = 1
    max_attempt = 4
    
    while attempt <= max_attempt:
        res = None

        try:
            if isinstance(body, dict) or isinstance(body, list):
                if debug_api:
                    print("get_data: sending json")
                res = await getattr(session, method.lower())(
                    url=url,
                    headers=headers,
                    json=body,
                    params=params,
                    follow_redirects=is_follow_redirects,
                    timeout = timeout
                )

            elif body:
                if debug_api:
                    print("get_data: sending data")

                res = await getattr(session, method.lower())(
                    url=url,
                    headers=headers,
                    data=body,
                    params=params,
                    follow_redirects=is_follow_redirects,
                    timeout = timeout
                )

            else:
                if debug_api:
                    print("get_data: no body")

                res = await getattr(session, method.lower())(
                    url=url,
                    headers=headers,
                    params=params,
                    follow_redirects=is_follow_redirects,
                    timeout = timeout
                )

            if debug_api:
                print("get_data_response", res)

            if return_raw:
                return res

            return rgd.ResponseGetData._from_httpx_response(res, auth=auth)

        except Exception as e:
            print(f"ℹ️ get_data error - {e} at {url}")
            attempt +=1

            if attempt == max_attempt:
                raise GetData_Error(url=url, message=e)


        finally:
            if is_close_session:
                await session.aclose()



# %% ../../nbs/client/10_get_data.ipynb 10
class LooperError(Exception):
    def __init__(self, loop_stage: str, message):

        super().__init__(f"{loop_stage} - {message}")

# %% ../../nbs/client/10_get_data.ipynb 11
async def looper(
    auth: dmda.DomoAuth,
    session: httpx.AsyncClient,
    url,
    offset_params,
    arr_fn: callable,
    loop_until_end: bool = False,
    method="POST",
    body: dict = None,
    fixed_params: dict = None,
    offset_params_in_body: bool = False,
    body_fn=None,
    limit=1000,
    skip=0,
    maximum=2000,
    debug_api: bool = False,
    debug_loop: bool = False,
) -> rgd.ResponseGetData:

    maximum = maximum or 0

    is_close_session = False

    if not session:
        session = httpx.AsyncClient()
        is_close_session = True

    allRows = []
    isLoop = True

    res = None

    if maximum < limit and not loop_until_end:
        limit = maximum

    while isLoop:
        params = fixed_params or {}

        if offset_params_in_body:
            body[offset_params.get("offset")] = skip
            body[offset_params.get("limit")] = limit

        else:
            params[offset_params.get("offset")] = skip
            params[offset_params.get("limit")] = limit

        if body_fn:
            try:
                body = body_fn(skip, limit)

            except Exception as e:
                await session.aclose()
                raise LooperError(
                    loop_stage="processing body_fn", message=str(e)
                ) from e

        if debug_loop:
            print(f"\n🚀 Retrieving records {skip} through {skip + limit} via {url}")
            # pprint(params)

        res = await get_data(
            auth=auth,
            url=url,
            method=method,
            params=params,
            session=session,
            body=body,
            debug_api=debug_api,
        )

        if not res.is_success:
            if is_close_session:
                await session.aclose()
            return res

        try:
            newRecords = arr_fn(res)

        except Exception as e:
            await session.aclose()
            raise LooperError(loop_stage="processing arr_fn", message=str(e)) from e

        allRows += newRecords

        if loop_until_end and len(newRecords) != 0:
            maximum = maximum + limit

        if debug_loop:
            print({"all_rows": len(allRows), "new_records": len(newRecords)})

        if len(allRows) >= maximum or len(newRecords) == 0:
            if debug_loop:
                print(
                    f"\n🎉 Success - {len(allRows)} records retrieved from {url} in query looper\n"
                )

            break

        skip += len(newRecords)

        if skip + limit > maximum:
            limit = maximum - len(allRows)

        if debug_loop:
            print(f"skip: {skip}, limit: {limit}")

    if is_close_session:
        await session.aclose()

    return await rgd.ResponseGetData._from_looper(res=res, array=allRows)

# %% ../../nbs/client/10_get_data.ipynb 15
async def looper_aiohttp(
    auth: dmda.DomoAuth,
    session: aiohttp.ClientSession,
    url,
    offset_params,
    arr_fn: callable,
    loop_until_end: bool = False,
    method="POST",
    body: dict = None,
    fixed_params: dict = None,
    offset_params_in_body: bool = False,
    body_fn=None,
    limit=1000,
    skip=0,
    maximum=2000,
    debug_api: bool = False,
    debug_loop: bool = False,
) -> rgd.ResponseGetData:

    is_close_session = False

    if not session:
        session = aiohttp.ClientSession()
        is_close_session = True

    allRows = []
    isLoop = True

    res = None

    if maximum < limit:
        limit = maximum

    while isLoop:
        params = fixed_params or {}

        if offset_params_in_body:
            body[offset_params.get("offset")] = skip
            body[offset_params.get("limit")] = limit

        else:
            params[offset_params.get("offset")] = skip
            params[offset_params.get("limit")] = limit

        if body_fn:
            try:
                body = body_fn(skip, limit)

            except Exception as e:
                await session.aclose()
                raise LooperError(
                    loop_stage="processing body_fn", message=str(e)
                ) from e

        if debug_loop:
            print(f"\n🚀 Retrieving records {skip} through {skip + limit} via {url}")
            # pprint(params)

        res = await get_data_aiohttp(
            auth=auth,
            url=url,
            method=method,
            params=params,
            session=session,
            body=body,
            debug_api=debug_api,
        )

        if not res.is_success:
            if is_close_session:
                await session.aclose()
            return res

        try:
            newRecords = arr_fn(res)

        except Exception as e:
            await session.close()
            raise LooperError(loop_stage="processing arr_fn", message=str(e)) from e

        allRows += newRecords

        if loop_until_end and len(newRecords) != 0:
            maximum = maximum + limit

        if debug_loop:
            print({"all_rows": len(allRows), "new_records": len(newRecords)})

        if len(allRows) >= maximum or len(newRecords) == 0:
            if debug_loop:
                print(
                    f"\n🎉 Success - {len(allRows)} records retrieved from {url} in query looper\n"
                )

            break

        skip += len(newRecords)

        if skip + limit > maximum:
            limit = maximum - len(allRows)

            if debug_loop:
                print(f"skip: {skip}, limit: {limit}")

    if is_close_session:
        await session.close()

    return await rgd.ResponseGetData._from_looper(res=res, array=allRows)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoPublish.ipynb.

# %% auto 0
__all__ = ['DomoPublication_Subscription', 'DomoPublications', 'DomoPublication_Content', 'DomoPublication_UnexpectedContentType',
           'DomoPublication']

# %% ../../nbs/classes/50_DomoPublish.ipynb 2
from dataclasses import dataclass, field

from typing import Optional

import datetime as dt
import asyncio
import httpx

from fastcore.basics import patch_to

# import importlib
# import json
# import uuid
# import time

import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de
import domolibrary.routes.publish as publish_routes

# import Library.DomoClasses.DomoDataset as dmda
# import Library.DomoClasses.DomoLineage as dmdl

# %% ../../nbs/classes/50_DomoPublish.ipynb 4
@dataclass
class DomoPublication_Subscription:
    subscription_id: str
    publication_id: str
    domain: str
    created_dt: Optional[dt.datetime] = None

    @classmethod
    def _from_json(cls, json):

        dd = json
        if not isinstance(dd, util_dd.DictDot):
            dd = util_dd.DictDot(json)

        return cls(
            subscription_id=dd.id,
            publication_id=dd.publicationId,
            domain=dd.domain,
            created_dt=dt.datetime.fromtimestamp(dd.created / 1000)
            if dd.created
            else None,
        )




# %% ../../nbs/classes/50_DomoPublish.ipynb 6
class DomoPublications:
    auth : dmda.DomoAuth

# %% ../../nbs/classes/50_DomoPublish.ipynb 7
@dataclass
class DomoPublication_Content:
    content_id: str
    entity_type: str
    entity_id: str
    entity_domain: str
    is_v2: bool
    is_direct_content: bool

    @classmethod
    def _from_json(cls, obj: dict):

        dd = obj
        if not isinstance(dd, util_dd.DictDot):
            dd = util_dd.DictDot(obj)

        dmpc = cls(
            content_id=dd.id,
            entity_type=dd.content.type,
            entity_id=dd.content.domoObjectId,
            entity_domain=dd.content.domain,
            is_v2=dd.isV2,
            is_direct_content=dd.useDirectContent,
        )

        return dmpc

    def to_api_json(self):
        temp_dict = {
            "domain": self.entity_domain,
            "domoObjectId": self.entity_id,
            # "customerId": self.entity_domain,
            "type": self.entity_type,
        }
        return temp_dict


# %% ../../nbs/classes/50_DomoPublish.ipynb 9
class DomoPublication_UnexpectedContentType(Exception):
    def __init__(self, publication_id, content_type, domo_instance):
        super().__init__(f"DomoPublication_Instantiation: Unexpected content type {content_type} in publication {publication_id} in {domo_instance}")

# %% ../../nbs/classes/50_DomoPublish.ipynb 10
@dataclass
class DomoPublication:
    id: str
    name: str
    description: str
    is_v2: bool
    created_dt: dt.datetime

    auth: dmda.DomoAuth = field(default=None, repr=False)

    subscription_authorizations: [DomoPublication_Subscription] = field(
        default_factory=list
    )
    content: [DomoPublication_Content] = field(default_factory=list)

    content_page_id_ls: [str] = field(default_factory=list)
    content_dataset_id_ls: [str] = field(default_factory=list)

    # lineage: dmdl.DomoLineage = None

    # def __post_init__(self):
    #     self.lineage = dmdl.DomoLineage(id=self.id,
    #                                     parent=self)

    @classmethod
    def _from_json(cls, obj, auth: dmda.DomoAuth = None):

        dd = util_dd.DictDot(obj)

        domo_pub = cls(
            id=dd.id,
            name=dd.name,
            description=dd.description,
            created_dt=dt.datetime.fromtimestamp(dd.created / 1000)
            if dd.created
            else None,
            is_v2=dd.isV2,
            auth=auth,
        )

        if dd.subscriptionAuthorizations and len(dd.subscriptionAuthorizations) > 0:
            domo_pub.subscription_authorizations = [
                DomoPublication_Subscription._from_json(sub)
                for sub in dd.subscriptionAuthorizations
            ]

        # publish only supports sharing pages and datasets
        if dd.children and len(dd.children) > 0:
            for child in dd.children:

                dmpc = DomoPublication_Content._from_json(child)
                domo_pub.content.append(dmpc)

                if dmpc.entity_type == "PAGE":
                    domo_pub.content_page_id_ls.append(dmpc.entity_id)

                elif dmpc.entity_type == "DATASET":
                    domo_pub.content_dataset_id_ls.append(dmpc.entity_id)

                else:
                    raise DomoPublication_UnexpectedContentType(
                        publication_id=domo_pub.id, 
                        content_type=dmpc.entity_type, 
                        domo_instance=auth.domo_instance)

        return domo_pub


# %% ../../nbs/classes/50_DomoPublish.ipynb 20
@patch_to(DomoPublications, cls_method=True)
async def search_publications(cls: DomoPublications,
                              auth = dmda.DomoAuth,
                              search_term: str = None,
                              session: httpx.AsyncClient = None,
                              debug_api: bool = False,
                              return_raw: bool = False):

    res = await publish_routes.search_publications(auth=auth)

    if return_raw:
        return res

    if not res.is_success or (res.is_success and len(res.response) == 0):
        return None

    return [DomoPublication._from_json(sub_obj)for sub_obj in res.response]




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoGrant.ipynb.

# %% auto 0
__all__ = ['DomoGrant']

# %% ../../nbs/classes/50_DomoGrant.ipynb 2
from dataclasses import dataclass, field

import domolibrary.client.DomoAuth as dmda
import domolibrary.utils.DictDot as util_dd
import domolibrary.routes.grant as grant_routes


# %% ../../nbs/classes/50_DomoGrant.ipynb 3
@dataclass
class DomoGrant:
    id: str
    display_group: str
    title: str
    depends_on_ls: list[str]
    description: str = None
    role_membership_ls: list[str] = field(default=None)

    @classmethod
    def _from_json(cls, obj):

        dd = obj
        if not isinstance(dd, util_dd.DictDot):
            dd = util_dd.DictDot(obj)

        return cls(id=dd.authority,
                   display_group=dd.authorityUIGroup,
                   depends_on_ls=dd.dependsOnAuthorities,
                   title=dd.title,
                   description=dd.description,
                   role_membership_ls=[str(role) for role in dd.roleIds])




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoUser.ipynb.

# %% auto 0
__all__ = ['DomoUser', 'DomoUsers']

# %% ../../nbs/classes/50_DomoUser.ipynb 3
from fastcore.basics import patch_to


# %% ../../nbs/classes/50_DomoUser.ipynb 4
from dataclasses import dataclass, field
from typing import Optional
import httpx

from pprint import pprint

import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.Logger as lc
import domolibrary.client.DomoError as de
import domolibrary.routes.user as user_routes

# %% ../../nbs/classes/50_DomoUser.ipynb 6
@dataclass
class DomoUser:
    """a class for interacting with a Domo User"""

    id: str
    display_name: str = None
    email_address: str = None
    role_id: str = None

    publisher_domain: str = None
    subscriber_domain: str = None
    virtual_user_id: str = None

    auth: Optional[dmda.DomoAuth] = field(repr=False, default=None)

    @classmethod
    def _from_search_json(cls, auth, user_obj):
        user_dd = util_dd.DictDot(user_obj)

        return cls(
            auth=auth,
            id=str(user_dd.id) or str(user_dd.userId),
            display_name=user_dd.displayName,
            email_address=user_dd.emailAddress,
            role_id=user_dd.roleId,
        )

    @classmethod
    def _from_virtual_json(cls, auth, user_obj):
        user_dd = util_dd.DictDot(user_obj)

        return cls(
            id=user_dd.id,
            auth=auth,
            publisher_domain=user_dd.publisherDomain,
            subscriber_domain=user_dd.subscriberDomain,
            virtual_user_id=user_dd.virtualUserId,
        )

    @classmethod
    def _from_bootstrap_json(cls, auth, user_obj):

        dd = user_obj
        if isinstance(user_obj, dict):
            dd = util_dd.DictDot(user_obj)

        return cls(id=dd.id, display_name=dd.displayName, auth=auth)

# %% ../../nbs/classes/50_DomoUser.ipynb 8
@patch_to(DomoUser)
async def reset_password(self: DomoUser, new_password: str, debug_api: bool = False):
    """reset your password, will respect password restrictions set up in the Domo UI"""

    res = await user_routes.reset_password(
        auth=self.auth, user_id=self.id, new_password=new_password, debug_api=debug_api
    )

    return res

# %% ../../nbs/classes/50_DomoUser.ipynb 9
@patch_to(DomoUser, cls_method=True)
async def request_password_reset(
    cls,
    domo_instance: str,
    email: str,
    locale: str = "en-us",
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    """request password reset email.  Note: does not require authentication."""

    return await user_routes.request_password_reset(
        domo_instance=domo_instance,
        email=email,
        locale=locale,
        debug_api=debug_api,
        session=session,
    )

# %% ../../nbs/classes/50_DomoUser.ipynb 10
@patch_to(DomoUser, cls_method=True)
async def create_user(
    cls: DomoUser,
    auth: dmda.DomoAuth,
    display_name,
    email,
    role_id,
    password: str = None,
    send_password_reset_email: bool = False,
    debug_api: bool = False,
):
    """class method that creates a new Domo user"""

    res = await user_routes.create_user(
        auth=auth,
        display_name=display_name,
        email=email,
        role_id=role_id,
        debug_api=debug_api,
    )

    if res.status != 200:
        return None

    dd = util_dd.DictDot(res.response)
    u = cls(
        domo_instance=auth.domo_instance,
        auth=auth,
        id=dd.id or dd.userId,
        display_name=dd.displayName,
        email_address=dd.emailAddress,
    )

    if password:
        await u.reset_password(new_password=password)

    if send_password_reset_email:
        await u.request_password_reset(
            domo_instance=auth.domo_instance, email=u.email_address
        )

    return u


@patch_to(DomoUser)
async def set_user_landing_page(
    self: DomoUser,
    page_id: str,
    user_id: str = None,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
):

    res = await user_routes.set_user_landing_page(
        auth=auth or self.auth,
        page_id=page_id,
        user_id=self.id or user_id,
        debug_api=debug_api,
    )

    if res.status != 200:
        return False

    return True

# %% ../../nbs/classes/50_DomoUser.ipynb 12
@dataclass
class DomoUsers:
    """a class for searching for Users"""

    logger: Optional[lc.Logger] = None

    @classmethod
    def _users_to_domo_user(cls, user_ls, auth: dmda.DomoAuth):
        return [
            DomoUser._from_search_json(auth=auth, user_obj=user_obj)
            for user_obj in user_ls
        ]

    @classmethod
    def _users_to_virtual_user(cls, user_ls, auth: dmda.DomoAuth):
        return [
            DomoUser._from_virtual_json(auth=auth, user_obj=user_obj)
            for user_obj in user_ls
        ]

    def _generate_logger(self, logger: Optional[lc.Logger] = None):
        self.logger = logger or self.logger or lc.loger()

# %% ../../nbs/classes/50_DomoUser.ipynb 13
@patch_to(DomoUsers, cls_method=True)
async def all_users(
    cls: DomoUsers,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    debug_prn: bool = False,
    debug_log: bool = False,
    logger: Optional[lc.Logger] = None,
) -> [DomoUser]:
    """retrieves all users from Domo"""

    logger = logger or lc.Logger(app_name="all_users")

    res = await user_routes.get_all_users(auth=auth, debug_api=debug_api)

    if not res.is_success:
        return None

    users_ls = res.response

    message = f"{len(users_ls)} users retrieved from {auth.domo_instance}"

    if debug_prn:
        print(message)
    logger.log_info(message=message, debug_log=debug_log)

    return cls._users_to_domo_user(user_ls=users_ls, auth=auth)


# %% ../../nbs/classes/50_DomoUser.ipynb 16
@patch_to(DomoUsers, cls_method=True)
async def by_id(
    cls: DomoUsers,
    user_ids: list[str],  # can search for one or multiple users
    auth: dmda.DomoAuth,
    only_allow_one: bool = True,
    debug_api: bool = False,
    return_raw: bool = False,
) -> list:

    body = user_routes.generate_search_users_body_by_id(user_ids)

    try:
        res = await user_routes.search_users(
            return_raw=False,
            body=body,
            debug_api=debug_api,
            auth=auth,
        )

    except user_routes.SearchUser_NoResults as e:
        print(e)
        return None

    if return_raw:
        return res

    domo_users = cls._users_to_domo_user(user_ls=res.response, auth=auth)

    if only_allow_one:
        return domo_users[0]

    return domo_users

# %% ../../nbs/classes/50_DomoUser.ipynb 20
@patch_to(DomoUsers, cls_method=True)
def util_match_domo_users_to_emails(
    cls: DomoUsers, domo_users: list[DomoUser], user_email_ls: list[str]
) -> list:
    """pass in an array of user emails to match against an array of Domo User"""

    matches = []
    for idx, email in enumerate(user_email_ls):
        match_user = next(
            (
                domo_user
                for domo_user in domo_users
                if email.lower() == domo_user.email_address.lower()
            ),
            None,
        )
        if match_user:
            matches.append(match_user)
    return matches


@patch_to(DomoUsers, cls_method=True)
def util_match_users_obj_to_emails(
    cls: DomoUsers, user_ls: list[dict], user_email_ls: list[str]
) -> list:
    """pass in an array of user emails to match against an array of Domo User"""

    matches = []
    for idx, email in enumerate(user_email_ls):
        match_user = next(
            (
                user_obj
                for user_obj in user_ls
                if email.lower() == user_obj.get("emailAddress").lower()
            ),
            None,
        )
        if match_user:
            matches.append(match_user)
    return matches


@patch_to(DomoUsers, cls_method=True)
async def by_email(
    cls: DomoUsers,
    user_email_ls: list[str],
    auth: dmda.DomoAuth,
    only_allow_one: bool = True,
    debug_api: bool = False,
    debug_prn: bool = False,
    return_raw: bool = False,
) -> list:

    body = user_routes.generate_search_users_body_by_email(user_email_ls=user_email_ls)

    if debug_prn:
        pprint(body)

    try:
        res = await user_routes.search_users(
            body=body, auth=auth, return_raw=False, debug_api=debug_api
        )

    except user_routes.SearchUser_NoResults as e:
        print(e)
        return None

    if return_raw:
        if only_allow_one:
            res.response = cls.util_match_users_obj_to_emails(
                res.response.get("users"), user_email_ls
            )[0]
        return res

    domo_users = cls._users_to_domo_user(res.response, auth=auth)

    if only_allow_one:
        return cls.util_match_domo_users_to_emails(domo_users, user_email_ls)[0]

    return domo_users

# %% ../../nbs/classes/50_DomoUser.ipynb 23
@patch_to(DomoUsers, cls_method=True)
async def virtual_user_by_subscriber_instance(
    cls: DomoUsers,
    subscriber_instance_ls: str,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    return_raw: bool = False,
):
    res = await user_routes.search_virtual_user_by_subscriber_instance(
        auth=auth,
        subscriber_instance_ls=subscriber_instance_ls,
        debug_api=debug_api,
    )

    if return_raw:
        return res

    if not res.is_success:
        return None

    user_ls = res.response

    domo_users = cls._users_to_virtual_user(user_ls, auth=auth)
    return domo_users[0]




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoRole.ipynb.

# %% auto 0
__all__ = ['DomoRole', 'DomoRoles', 'SearchRole_NotFound']

# %% ../../nbs/classes/50_DomoRole.ipynb 2
from dataclasses import dataclass, field
import httpx

from fastcore.basics import patch_to

import domolibrary.client.DomoAuth as dmda

import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoError as de
import domolibrary.routes.role as role_routes


# %% ../../nbs/classes/50_DomoRole.ipynb 3
@dataclass
class DomoRole:
    auth: dmda.DomoAuth = field(repr = False)

    id: str
    name: str = None
    description: str = None
    is_system_role: int = None

    grant_ls: list[str] = field(default_factory=list)
    membership_ls: list = field(default_factory=list)

    def __post_init__(self):
        self.is_system_role = 1 if self.id <= 5 else 0

    # @classmethod
    # def _from_str(cls, id, name, description=None, auth: DomoAuth = None):

    #     return cls(id=id,
    #             name=name,
    #             description=description,
    #             auth=auth
    #             )

    @classmethod
    def _from_json(cls, obj, auth=dmda.DomoAuth):

        dd = obj
        if not isinstance(dd, util_dd.DictDot):
            dd = util_dd.DictDot(obj)

        return cls(id=dd.id,
                   name=dd.name,
                   description=dd.description,
                   auth=auth
                   )


# %% ../../nbs/classes/50_DomoRole.ipynb 4
@patch_to(DomoRole)
async def get_role_grants(self: DomoRole,
                          auth: dmda.DomoAuth = None,
                          role_id: str = None,
                          debug_api: bool = False,
                          session: httpx.AsyncClient = None,
                          return_raw: bool = False
                          ):

    auth = auth or self.auth
    role_id = role_id or self.id
    res = await role_routes.get_role_grants(auth=auth,
                                            role_id=role_id,
                                            debug_api=debug_api,
                                            session=session)

    if return_raw:
        return res
        
    self.grant_ls = res.response

    return self.grant_ls


# %% ../../nbs/classes/50_DomoRole.ipynb 7
@dataclass
class DomoRoles:
    auth: dmda.DomoAuth


# %% ../../nbs/classes/50_DomoRole.ipynb 8
@patch_to(DomoRoles, cls_method=True)
async def get_roles(cls: DomoRoles,
                    auth: dmda.DomoAuth,
                    debug_api: bool = False,
                    session: httpx.AsyncClient = None,
                    return_raw: bool = False
                    ):

    res = await role_routes.get_roles(auth=auth, session=session, debug_api=debug_api)

    if return_raw:
        return res

    role_ls = res.response

    return [ DomoRole._from_json(role, auth) for role in res.response]


# %% ../../nbs/classes/50_DomoRole.ipynb 10
class SearchRole_NotFound(de.DomoError):
    def __init__(self, domo_instance,
                 role_id,
                 message='not found',
                 function_name='search_role'):

        super().__init__(domo_instance=domo_instance, message=message,
                         entity_id=role_id, function_name=function_name)

@patch_to(DomoRoles, cls_method=True)
async def search_role(cls: DomoRoles,
                    role_name: str,
                    auth: dmda.DomoAuth,
                    debug_api: bool = False,
                    session: httpx.AsyncClient = None,
                    return_raw: bool = False
                    ):

    all_roles = await DomoRoles.get_roles( auth=auth)
    
    domo_role = next(
        (role for role in all_roles if role.name == role_name), None)

    if not domo_role:
        raise SearchRole_NotFound(domo_instance=auth.domo_instance, role_id=role_name)

    return domo_role

# %% ../../nbs/classes/50_DomoRole.ipynb 17
@patch_to(DomoRoles, cls_method=True)
async def upsert_role(cls: DomoRoles,
                      auth: dmda.DomoAuth,
                      name: str,
                      description: str = None,
                      debug_api: bool = False,
                      debug_prn: bool = False,
                      session : httpx.AsyncClient=None
                      ):

    domo_role = await DomoRoles.search_role(role_name=name,
                                            auth = auth ,
                                            debug_api=debug_api,
                                            session=session,
                                            )

    if domo_role:
        return domo_role

    if debug_prn:
        print(
            f'No role match -- creating new role in {full_auth.domo_instance}')

    return await DomoRoles.create_role(auth=auth,
                                       name=name, description=description,
                                       debug_api=debug_api, session=session)




from types import SimpleNamespace


class DictDot(SimpleNamespace):
    def __init__(self, dictionary, **kwargs):
        super().__init__(**kwargs)

        for key, value in dictionary.items():
            if isinstance(value, dict):
                self.__setattr__(key, DictDot(value))
            elif isinstance(value, list):
                new_list = []
                for item in value:
                    if isinstance(item, dict):
                        new_list.append(DictDot(item))
                    else:
                        new_list.append(item)
                self.__setattr__(key, new_list)
            else:
                self.__setattr__(key, value)

    def __getattr__(self, item):
        return None



import aiohttp

from .get_data import get_data, looper
from ..DomoAuth import DomoFullAuth
from ...utils.ResponseGetData import ResponseGetData


async def get_whitelist(full_auth: DomoFullAuth,
                        session: aiohttp.ClientSession = None,
                        debug: bool = False,
                        log_results: bool = False
                        ) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/admin/companysettings/whitelist'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='GET',
        session=session,
        debug=False
    )

    return res


async def update_whitelist(full_auth: DomoFullAuth,
                           ip_address_list: list[str],
                           debug: bool = False,
                           log_results: bool = False
                           ) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/admin/companysettings/whitelist'

    body = {'addresses': ip_address_list}

    if debug:
        print(url)
        print(body)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        body=body,
        log_results=log_results,
        debug=debug
    )

    return res


async def update_authorized_domains(full_auth: DomoFullAuth, authorized_domain_list: list[str],
                                    debug: bool = False, log_results: bool = False,
                                    session: aiohttp.ClientSession = None):
    url = f'https://{full_auth.domo_instance}.domo.com/api/content/v1/customer-states/authorized-domains'

    body = {
        "name": "authorized-domains",
        "value": ",".join(authorized_domain_list)
    }

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        body=body,
        log_results=log_results,
        debug=debug,
        session=session
    )

    return res


async def get_authorized_domains(full_auth: DomoFullAuth,
                                 debug: bool = False, log_results: bool = False,
                                 session: aiohttp.ClientSession = None
                                 ):
    url = f'https://{full_auth.domo_instance}.domo.com/api/content/v1/customer-states/authorized-domains'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='GET',
        log_results=log_results,
        debug=debug,
        session=session
    )

    return res



from dataclasses import dataclass

from pprint import pprint

import aiohttp
import datetime as dt
import json

from Library.utils.DictDot import DictDot
from ..utils.DictDot import DictDot
from .DomoAuth import DomoFullAuth
from enum import Enum
from .routes import job_routes

class WatchDogType(Enum):
    DATASET_ERRORS = 'error_detection'
    DATASET_INDEX_TIME = 'max_indexing_time'
    LAST_UPDATED_DATA = 'last_data_updated'
    ROW_COUNT_CHANGE = 'row_count_variance'
    CUSTOM_QUERY = 'custom_query'
    OUTLIER_EXECUTION = 'execution_variance'

    
@dataclass
class DomoTrigger_Schedule:
    schedule_text: str = None
    schedule_type: str = 'scheduleTriggered'

    minute: int = None
    hour: int = None
    minute_str: str = None
    hour_str : str = None
        
    @classmethod
    def _from_str(cls, s_text, s_type):
        sched = cls(
            schedule_type=s_type,
            schedule_text=s_text)

        try:
            parsed_hour =s_text.split(' ')[2]
            parsed_minute=s_text.split(' ')[1]
            
            if "*" in parsed_hour or "/" in parsed_hour:
                sched.hour_str = parsed_hour
            else:
                sched.hour = int(float(parsed_hour))
            if "*" in parsed_minute:
                sched.minute_str = parsed_minute
            else:
                sched.minute = int(float(parsed_minute))

            return sched

        except Exception as e:
            print(f"unable to parse schedule {s_text}")
            print(e)

    def to_obj(self):
        return {'hour': int(self.hour),
                'minute': int(self.minute)}

    def to_schedule_obj(self):
        minute = self.minute_str if self.minute_str is not None else str(self.minute)
        hour = self.hour_str if self.hour_str is not None else str(self.hour)
        return {
            "eventEntity": f'0 {minute} {hour} ? * *',
            #old value on Jan 13
            #"eventEntity": f'0 {minute} {hour} 1/1 * ? *',
            "eventType": self.schedule_type
        }


@dataclass
class DomoTrigger:
    id: str
    job_id: str
    schedule: [DomoTrigger_Schedule] = None


@dataclass
class DomoJob:
    id: str
    name: str
    description :str
    remote_instance: str
    user_id: str
    application_id: str
    customer_id: str
    execution_timeout: int
    entity_ids: list
    job_type: str
    entity_type: str
    max_indexing_time_min :int
    variance_percent: int
    min_update_frequency_min: int
    sql_query:str
    notify_user_ids: list
    metrics_dataset_id: str
    notify_group_ids: list
    notify_email_addressess: list
    resources_requests: str
    resources_limits: str
    
    triggers: [DomoTrigger] = None

    @classmethod
    def _from_json(cls, obj):
        dd = DictDot(obj)
        print (dd.jobName)
        #print (dd)
        triggers_ls = obj.get('triggers', None)

        triggers_dj = [DomoTrigger(
            id=tg.get('triggerId'),
            job_id=tg.get('jobId'),

            schedule=DomoTrigger_Schedule._from_str(
                s_text=tg.get('eventEntity'),
                s_type=tg.get('eventType'))
        ) for tg in triggers_ls]

        return cls(id=dd.jobId,
                   application_id=dd.applicationId,
                   customer_id=dd.customerId,
                   name=dd.jobName,
                   description = dd.jobDescription,
                   user_id=dd.userId,
                   remote_instance=dd.executionPayload.remoteInstance.replace(
                       '.domo.com', '') if dd.executionPayload.remoteInstance else None,
                   execution_timeout = dd.executionTimeout,
                   entity_ids = dd.executionPayload.watcherParameters.entityIds if dd.executionPayload.watcherParameters else [],
                   job_type = dd.executionPayload.watcherParameters.type if dd.executionPayload.watcherParameters else [],
                   entity_type=dd.executionPayload.watcherParameters.entityType if dd.executionPayload.watcherParameters else [],
                   max_indexing_time_min = dd.executionPayload.watcherParameters.maxIndexingTimeInMinutes if dd.executionPayload.watcherParameters else [],
                   variance_percent = dd.executionPayload.watcherParameters.variancePercent if dd.executionPayload.watcherParameters else [],
                   min_update_frequency_min = dd.executionPayload.watcherParameters.minDataUpdateFrequencyInMinutes if dd.executionPayload.watcherParameters else [],
                   sql_query=dd.executionPayload.watcherParameters.sqlQuery if dd.executionPayload.watcherParameters else [],
                   notify_user_ids=dd.executionPayload.notifyUserIds,
                   metrics_dataset_id=dd.executionPayload.metricsDatasetId,
                   notify_group_ids=dd.executionPayload.notifyGroupIds,
                   notify_email_addressess=dd.executionPayload.notifyEmailAddresses,
                   resources_requests =dd.resources.requests.memory,
                   resources_limits = dd.resources.limits.memory,
                   triggers=triggers_dj)
    def get_body (self):
        body = {
        "jobName": self.name,
        "jobDescription": self.description,
        "executionTimeout": self.execution_timeout,
        "accounts": [],
        "executionPayload": {
            "notifyUserIds": self.notify_user_ids or [],
            "notifyGroupIds": self.notify_group_ids or [],
            "notifyEmailAddresses": self.notify_email_addressess or [],
        "watcherParameters": {
          "entityIds": self.entity_ids,
          "type": self.job_type,
          "entityType": self.entity_type,
          "maxIndexingTimeInMinutes":self.max_indexing_time_min,
          "variancePercent": self.variance_percent,
          "sqlQuery": self.sql_query
            },
        "metricsDatasetId": self.metrics_dataset_id
          },
        "resources": {
            "requests": {"memory": self.resources_requests},
            "limits": {"memory": self.resources_limits}
        },
        "triggers": [self.triggers[0].schedule.to_schedule_obj()] if len(self.triggers)>0 else []
        }
        return body

        
    @classmethod
    async def create_domostats_job(cls,
                                   full_auth: DomoFullAuth,
                                   domostats_schedule: DomoTrigger_Schedule,
                                   application_id: str,
                                   target_instance: str,
                                   report_dict: dict,
                                   output_dataset_id: str,
                                   account_id: str,
                                   execution_timeout: int = 1440,
                                   debug: bool = False, log_results: bool = False,
                                   session: aiohttp.ClientSession = None):

        schedule_obj = domostats_schedule.to_schedule_obj()

        body = job_routes.generate_body_remote_domostats(target_instance=target_instance,
                                                         report_dict=report_dict,
                                                         output_dataset_id=output_dataset_id,
                                                         account_id=account_id,
                                                         schedule_ls=[
                                                             schedule_obj],
                                                         execution_timeout=execution_timeout)
        
        res = await job_routes.add_job(full_auth=full_auth,
                                       application_id=application_id,
                                       body=body,
                                       debug=debug,
                                       log_results=log_results,
                                       session=session)
        if debug:
            print(res)

        if res.status != 200:
            return False

        return True

    @classmethod
    async def create_watchdog_job(cls,
                                   full_auth: DomoFullAuth,
                                   body : str,
                                   debug: bool = False, log_results: bool = False,
                                   session: aiohttp.ClientSession = None):
            
            
        res = await job_routes.add_job(full_auth=full_auth,
                                       application_id=application_id,
                                       body=body,
                                       debug=debug,
                                       log_results=log_results,
                                       session=session)

#         if debug:
#             print(res)

        if res.status != 200:
            return False

        return True
    
    @classmethod
    async def generate_watchdog_body(cls,
                                   watchdog_report_type : WatchDogType,
                                   watchdog_schedule: DomoTrigger_Schedule,
                                   job_name: str,
                                   notify_user_ids_ls: list,
                                   notify_group_ids_ls: list,
                                   notify_emails_ls: list,
                                   entity_ids_ls: list,
                                   entity_type : str,
                                   metric_dataset_id: str,
                                   sql_query :str = None,
                                   variance_percent: int = 30,
                                   max_indexing_time_mins: int = 30,
                                   execution_timeout: int = 1440,
                                   min_update_frequency_min: int = 1440,
                           ):
        schedule_obj = watchdog_schedule.to_schedule_obj()

        body = job_routes.generate_body_watchdog_generic(job_name=job_name,
                                                         notify_user_ids_ls=notify_user_ids_ls,
                                                         notify_group_ids_ls=notify_group_ids_ls,
                                                         notify_emails_ls=notify_emails_ls,
                                                         entity_ids_ls = entity_ids_ls,
                                                         entity_type=entity_type,
                                                         metric_dataset_id=metric_dataset_id,
                                                         schedule_ls=[
                                                             schedule_obj],
                                                         job_type = watchdog_report_type.value,
                                                         execution_timeout=execution_timeout)
        

        
        if watchdog_report_type == watchdog_report_type.DATASET_INDEX_TIME:
            child ={ "maxIndexingTimeInMinutes": max_indexing_time_mins }
            body["executionPayload"]["watcherParameters"].update(child)

  
        
        if watchdog_report_type == watchdog_report_type.ROW_COUNT_CHANGE or watchdog_report_type == watchdog_report_type.OUTLIER_EXECUTION:
            child ={ "variancePercent": variance_percent }
            body["executionPayload"]["watcherParameters"].update(child)
            
        
        if watchdog_report_type == watchdog_report_type.CUSTOM_QUERY:
            child ={ "sqlQuery": sql_query }
            body["executionPayload"]["watcherParameters"].update(child)
            
        if watchdog_report_type == watchdog_report_type.LAST_UPDATED_DATA:
            child ={ "minDataUpdateFrequencyInMinutes": min_update_frequency_min }
            body["executionPayload"]["watcherParameters"].update(child)
            
            
        return body
                            
                            
                            
    @classmethod
    async def update_job(cls,
                         full_auth: DomoFullAuth,
                         body: str,
                         job_id: str,
                         application_id: str,
                         debug: bool = False,
                         log_results: bool = False,
                        session: aiohttp.ClientSession = None):
        

        res = await job_routes.update_job(full_auth=full_auth,
                                          body=body,
                                       application_id=application_id,
                                       job_id = job_id,
                                       debug=debug,
                                       log_results=log_results,
                                       session=session)


        if res.status != 200:
            return False

        return True
                         


# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/instance_config.ipynb.

# %% auto 0
__all__ = ['get_allowlist', 'Allowlist_UnableToUpdate', 'set_allowlist']

# %% ../../nbs/routes/instance_config.ipynb 2
import httpx

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de


# %% ../../nbs/routes/instance_config.ipynb 3
async def get_allowlist(
    auth: dmda.DomoFullAuth,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    return_raw: bool = False,
) -> rgd.ResponseGetData:

    if auth.__class__.__name__ != "DomoFullAuth":
        raise dmda.InvalidAuthTypeError(
            function_name="get_allowlist",
            domo_instance=auth.domo_instance,
            required_auth_type=dmda.DomoFullAuth,
        )

    url = f"https://{auth.domo_instance}.domo.com/admin/companysettings/whitelist"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        headers={"accept": "*/*"},
        session=session,
        debug_api=debug_api,
        is_follow_redirects=True,
        return_raw=return_raw,
    )

    return res

# %% ../../nbs/routes/instance_config.ipynb 8
class Allowlist_UnableToUpdate(de.DomoError):
    def __init__(
        self, 
        status: int, reason: str, domo_instance:str, function_name: str = "update_allowlist"
    ):

        super().__init__(
            function_name=function_name,
            status=status,
            message=f"unable to update allowlist: {reason}",
            domo_instance=domo_instance,
        )

# %% ../../nbs/routes/instance_config.ipynb 9
async def set_allowlist(
    auth: dmda.DomoAuth,
    ip_address_ls: list[str],
    debug_api: bool = False,
    return_raw: bool = False,
    session: httpx.AsyncClient = None
) -> rgd.ResponseGetData:
    """companysettings/whitelist API only allows users to SET the allowlist does not allow INSERT or UPDATE"""

    url = f"https://{auth.domo_instance}.domo.com/admin/companysettings/whitelist"

    body = {"addresses": ip_address_ls}

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=body,
        debug_api=debug_api,
        is_follow_redirects=True,
        return_raw=return_raw,
        session = session,
        headers={"accept": "text/plain"},
    )
    if not res.is_success:
        raise Allowlist_UnableToUpdate(
            status=res.status, reason=res.response, domo_instance=auth.domo_instance
        )

    if res.is_success:
        res.response = f"allow list updated from {res.response}"

    return res



import aiohttp
from pprint import pprint

from ..DomoAuth import DomoAuth
from ...utils.ResponseGetData import ResponseGetData

from ast import literal_eval


async def get_data(url, method, auth: DomoAuth,
                   content_type=None,
                   headers=None,
                   session: aiohttp.ClientSession = None,
                   body=None,
                   params: dict = None,
                   log_results=False, debug=False) -> ResponseGetData:
    from pprint import pprint
    import aiohttp

    if auth and not auth.token:
        await auth.get_auth_token()

    if headers is None:
        headers = {}

    is_close_session = False
    if session is None:
        is_close_session = True
        session = session or aiohttp.ClientSession()

    headers = {
        'Content-Type': content_type or 'application/json',
        'Connection': 'keep-alive',
        'accept': 'application/json, text/plain',
        **headers
    }

    if auth:
        headers.update(**auth.auth_header)

    if debug:
        pprint({'method': method,
                'url': url,
                'headers': headers,
                'json': body,
                'params': params})

    # if params:
    #     params = {k: literal_eval(v) for k, v in params.items()}
    #     print(params)

    if headers.get('Content-Type') == 'application/json':
        if debug:
            print('passing json')

        res = await session.request(method=method.upper(),
                                    url=url, headers=headers,
                                    json=body,
                                    params=params)

    elif body is not None:
        res = await session.request(method=method.upper(),
                                    url=url,
                                    headers=headers,
                                    data=body,
                                    params=params)

    else:
        res = await session.request(method=method.upper(),
                                    url=url,
                                    headers=headers,
                                    params=params)

    if is_close_session:
        await session.close()

    if log_results:
        try:
            pprint({'url': url, 'text': await res.json()})
        except:
            pprint({'url': url, 'text': await res.text()})

    if res.status == 200 \
            and hasattr(res, 'headers') \
            and res.headers.get('Content-Type') \
            and 'application/json' in res.headers.get('Content-Type'):

        if debug:
            print('return json')

        try:
            response = await res.json()

            # if debug:
            #     print({'get_data_json_response': response})

            return ResponseGetData(status=res.status,
                                   response=response,
                                   is_success=True,
                                   auth=auth)
        except:
            if debug:
                print('SUCCESS - however. Expected JSON but received read io')
            response = await res.read()
            return ResponseGetData(status=res.status,
                                   response=response.decode(),
                                   is_success=True,
                                   auth=auth)
            raise Exception

    elif res.status == 200:
        if debug:
            print('return text')
        return ResponseGetData(res.status, await res.text(), is_success=True, auth=auth)

    else:
        return ResponseGetData(res.status, response={'success': False, 'reason': res.reason}, is_success=False,
                               auth=auth)


async def looper(auth: DomoAuth,
                 session: aiohttp.ClientSession,
                 url,
                 offset_params,
                 arr_fn: callable,
                 alter_maximum_fn: callable = None,
                 loop_until_end: bool = False,
                 method='POST',
                 body: dict = None,
                 fixed_params: dict = None,
                 offset_params_in_body: bool = False,
                 body_fn=None,
                 limit=1000,
                 maximum=2000,
                 debug: bool = False):
    allRows = []
    skip = 0
    isLoop = True

    while isLoop:
        params = fixed_params or {}

        if offset_params_in_body:
            body[offset_params.get('offset')] = skip
            body[offset_params.get('limit')] = limit

        else:
            params[offset_params.get('offset')] = skip
            params[offset_params.get('limit')] = limit

        if body_fn:
            body = body_fn(skip,
                           limit)

        if debug:
            pprint(params)
            print(
                f'Retrieving records {skip} through {skip + limit} via {url}')

        res = await get_data(auth=auth,
                             url=url,
                             method=method,
                             params=params,
                             session=session,
                             body=body,
                             debug=debug)

        newRecords = arr_fn(res)
        # print('loop', newRecords)
        # process rows
        allRows += newRecords

        if skip == 0 and alter_maximum_fn:
            maximum = alter_maximum_fn(res)
            #print(f'the new maximum is: {maximum}')

        if loop_until_end and len(newRecords) != 0:
            maximum = maximum + limit

        if debug:
            print(len(allRows), len(newRecords))

        if maximum <= len(allRows) or len(newRecords) < limit:
            if debug:
                print(
                    f'{len(allRows)} records retrieved from {url} in query looper\n')
            isLoop = False

        skip += limit

    return allRows



import os
from os.path import exists
from pprint import pprint

from dotenv import load_dotenv

from . import DictDot as dd


def split_str_to_obj(env_string: str, value_split_keys: list[str]):
    env_list = env_string.split('|')
    obj = {}

    for index, new_obj_key in enumerate(value_split_keys):
        obj[new_obj_key] = env_list[index]

    return dd.DictDot(obj)


def split_creds(env: dd.DictDot, key_starts_with: str, env_var_list: list[str]):
    env_lines_to_split = [getattr(env, key) for key in dir(
        env) if key.startswith(key_starts_with)]

    return [split_str_to_obj(env_string=line, value_split_keys=env_var_list) for line in env_lines_to_split]


def read_creds_from_dotenv(env_path: str = '.env',
                           params: list[str] = None,
                           debug: bool = False) -> dd.DictDot:
    """use_prod = false will replace all PROD values with matching TEST values"""

    file_exists = exists(env_path)
    if not file_exists:
        print(f"file not found at -- {env_path}")
        raise Exception('env file not found')

    load_dotenv(env_path)
    params = params or list(os.environ.keys())

    params_res = {}
    for param in params:
        param = str(param)
        params_res.update({param: os.environ.get(param)})

    if debug:
        pprint(vars(os.environ))
        pprint({'read_creds_from_params': params_res})

    return dd.DictDot(params_res)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoDatacenter.ipynb.

# %% auto 0
__all__ = ['DomoEntity', 'DomoDatacenter']

# %% ../../nbs/classes/50_DomoDatacenter.ipynb 3
import asyncio
import importlib
from dataclasses import dataclass, field
from enum import Enum

import httpx

from fastcore.basics import patch_to

import domolibrary.client.DomoAuth as dmda
import domolibrary.classes.DomoDataset as dmds

import domolibrary.routes.account as account_routes
import domolibrary.routes.datacenter as datacenter_routes

# %% ../../nbs/classes/50_DomoDatacenter.ipynb 4
class DomoEntity(Enum):
    "enumerates the entities that can be passed to the datacenter Search API"
    
    DATASET = "DATA_SOURCE" #dataset
    DATAFLOW = "DATAFLOW" #  dataflow
    PAGE = "PAGE" # page
    CARD = "CARD" # card
    ACCOUNT = "account" #account type


# %% ../../nbs/classes/50_DomoDatacenter.ipynb 5
class DomoDatacenter:
    "class for quering entities in the datacenter"
    auth: dmda.DomoAuth = None

    def __init__(self, auth):
        self.auth = auth

# %% ../../nbs/classes/50_DomoDatacenter.ipynb 6
@patch_to(DomoDatacenter, cls_method=True)
def generate_search_datacenter_body(
    cls,
    entity_name : str = None, # entity to search for, uses fuzzy logic
    entity_ls: list[str] = list[DomoEntity], # list of entity types to search
    filter_ls: [dict] = None, # filter parameters
    combineResults: bool = False, # facet / split results
    count: int = 100,
    offset: int = 0,
):
    body = {"query": "*" }
    if DomoEntity.ACCOUNT in entity_ls:
        body.update({
            "entityList": [[DomoEntity.ACCOUNT.value]],
        })

        if entity_name:
            body.update({"query": f"*{entity_name}*"})
    else:
        body.update({
            "entities": [entity.value for entity in entity_ls]
        })

        if entity_name:
            body.update({field: "name_sort", filterType: "wildcard", query: f"*{entity_name}*"})

    body.update({
        "count": count, "offset": offset,
        "filters": filter_ls or [],
        "combineResults": combineResults,
        
    })

    return body




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/role.ipynb.

# %% auto 0
__all__ = ['get_roles', 'get_role_grants', 'get_role_membership', 'create_role', 'set_default_role', 'update_role_grants',
           'role_membership_add_user']

# %% ../../nbs/routes/role.ipynb 2
import httpx

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de


# %% ../../nbs/routes/role.ipynb 5
async def get_roles(auth: dmda.DomoAuth,
                    debug_api: bool = False,
                    session: httpx.AsyncClient = None,
                    ) -> rgd.ResponseGetData:
    url = f'https://{auth.domo_instance}.domo.com/api/authorization/v1/roles'

    res = await gd.get_data(
        auth=auth,
        url=url,
        method='GET',
        debug_api=debug_api,
        session=session
    )

    if len(res.response) == 0 or not res.is_success:
        raise Role_NotRetrieved(
            domo_instance=auth.domo_instance,
            function_name='get_roles',
            status=res.status,
            message = 'failed to retrieve roles')
    
    return res


# %% ../../nbs/routes/role.ipynb 8
async def get_role_grants(auth: dmda.DomoAuth,
                          role_id: str,
                          debug_api: bool = False,
                          session : httpx.AsyncClient = None
                          ) -> rgd.ResponseGetData:
    url = f'https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/authorities'

    if debug_api:
        print(url)

    res = await gd.get_data(
        auth=auth,
        url=url,
        method='GET',
        debug_api=debug_api,
        session = session
    )

    if len(res.response) == 0:
        raise Role_NotRetrieved(
            domo_instance= auth.domo_instance,
            function_name='get_role_grants',
            message=f"no grants returned for role {role_id}",
            status=res.status
        )

    return res

# %% ../../nbs/routes/role.ipynb 11
async def get_role_membership(auth: dmda.DomoAuth,
                              role_id: str,
                              debug_api: bool = False, 
                              session: httpx.AsyncClient = None,
                              return_raw : bool = False
                              ) -> rgd.ResponseGetData:
    url = f'https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/users'

    if debug_api:
        print(url)

    res = await gd.get_data(
        auth=auth,
        url=url,
        method='GET',
        debug_api=debug_api,
        session=session
    )

    if len(res.response.get('users')) == 0:
        raise Role_NotRetrieved(
            domo_instance=auth.domo_instance,
            function_name='get_role_membership',
            message=f"no users returned for role {role_id}",
            status=res.status
        )

    if return_raw:
        return res
    
    res.response = res.response.get('users')
    
    return res

# %% ../../nbs/routes/role.ipynb 15
async def create_role(auth: dmda.DomoAuth,
                      name: str,
                      description: str,
                      debug_api: bool = False) -> rgd.ResponseGetData:
    url = f'https://{auth.domo_instance}.domo.com/api/authorization/v1/roles'

    body = {
        'name': name,
        'description': description
    }

    res = await gd.get_data(
        auth=auth,
        url=url,
        method='POST',
        debug_api=debug_api,
        body=body
    )

    return res

# %% ../../nbs/routes/role.ipynb 16
async def set_default_role(auth: dmda.DomoAuth,
                           role_id: str,
                           debug_api: bool = False,
                           session: httpx.AsyncClient = None) -> rgd.ResponseGetData:
    url = f'https://{auth.domo_instance}.domo.com/api/content/v1/customer-states/user.roleid.default'

    body = {
        'name': 'user.roleid.default',
        'value': role_id
    }

    res = await gd.get_data(
        auth=auth,
        url=url,
        method='PUT',
        debug_api=debug_api,
        body=body,
        session=session
    )
    return res


# %% ../../nbs/routes/role.ipynb 20
async def update_role_grants(auth: dmda.DomoAuth,
                             role_id: str,
                             role_grant_list: list[str],
                             debug_api: bool = False) -> rgd.ResponseGetData:
    url = f'https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/authorities'

    if debug_api:
        print(url)

    res = await gd.get_data(auth=auth,
                         url=url,
                         method='PUT',
                         debug_api=debug_api, body=role_grant_list)
    return res

# %% ../../nbs/routes/role.ipynb 21
async def role_membership_add_user(auth: dmda.DomoAuth,
                                   role_id: str,
                                   user_list: list[str],
                                   debug_api: bool = False) -> rgd.ResponseGetData:
    url = f'https://{auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/users'

    if debug_api:
        print(url)

    res = await gd.get_data(
        auth=auth,
        url=url,
        method='PUT',
        debug_api=debug_api,
        body=user_list
    )
    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/group.ipynb.

# %% auto 0
__all__ = ['generate_body_create_group', 'CreateGroup_Error', 'create_group', 'generate_body_update_group_membership']

# %% ../../nbs/routes/group.ipynb 2
import httpx

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de

# %% ../../nbs/routes/group.ipynb 3
def generate_body_create_group(group_name: str,
                               group_type: str = 'open',
                               description: str = '') -> dict:
    """ Generates the body to create group for content_v2_group API"""
    body = {"name": group_name, 
            "type": group_type,
            "description": description}

    return body

# %% ../../nbs/routes/group.ipynb 6
class CreateGroup_Error(de.DomoError):
    def __init__(self, status, message, domo_instance, function_name = "create_group"):
        super().__init__(function_name = function_name, status = status, message = message , domo_instance = domo_instance)

async def create_group(auth: dmda.DomoAuth,
                       group_name: str,
                       group_type: str = 'open',
                       description: str = '',
                       debug_api: bool = False,
                       session: httpx.AsyncClient = None
                       ) -> rgd.ResponseGetData:
    # body : {"name": "GROUP_NAME", "type": "open", "description": ""}

    body = generate_body_create_group(
        group_name=group_name, group_type=group_type, description=description)

    url = f'https://{auth.domo_instance}.domo.com/api/content/v2/groups/'

    res= await gd.get_data(
        auth=auth,
        url=url,
        method='POST',
        body=body,
        debug_api=debug_api,
        session = session
    )

    if not res.is_success:
        raise CreateGroup_Error(
            status = res.status, 
            message = res.response,
            domo_instance = auth.domo_instance, 
            function_name="create_group")

    return res


# %% ../../nbs/routes/group.ipynb 11
def generate_body_update_group_membership(group_id: str,
                                          add_user_arr: list[str] = None,
                                          remove_user_arr: list[str] = None,
                                          add_owner_user_arr: list[str] = None,
                                          remove_owner_user_arr: list[str] = None) -> list[dict]:
    body = {"groupId": int(group_id)}
    if add_owner_user_arr:
        body.update({"addOwners": [{"type": "USER", "id": str(
            userId)} for userId in add_owner_user_arr]})

    if remove_owner_user_arr:
        body.update({"removeOwners": [{"type": "USER", "id": str(
            userId)} for userId in remove_owner_user_arr]})

    if remove_user_arr:
        body.update({"removeMembers": [
                    {"type": "USER", "id": str(userId)} for userId in remove_user_arr]})
    if add_user_arr:
        body.update(
            {"addMembers": [{"type": "USER", "id": str(userId)} for userId in add_user_arr]})

    return [body]
